# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio.abstract_driver

import json
import socket
from enum import Enum


class ahioDriverInfo(ahio.abstract_driver.AbstractahioDriverInfo):
    NAME = 'GenericTCPIO'
    AVAILABLE = True


class Driver(ahio.abstract_driver.AbstractDriver):
    _socket = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._socket:
            self._socket.send(b'QUIT\n')
            self._socket.close()

    def setup(self, address, port):
        """Connects to server at `address`:`port`.

        Connects to a TCP server listening at `address`:`port` that implements
        the protocol described in the file "Generic TCP I:O Protocol.md"

        @arg address IP or address to connect to.
        @arg port port to connect to.

        @throw RuntimeError if connection was successiful but protocol isn't
               supported.
        @throw any exception thrown by `socket.socket`'s methods.
        """
        address = str(address)
        port = int(port)
        self._socket = socket.socket()
        self._socket.connect((address, port))
        self._socket.send(b'HELLO 1.0\n')
        with self._socket.makefile() as f:
            if f.readline().strip() != 'OK':
                raise RuntimeError('Protocol not supported')

    def __clamp(self, value, min, max):
        return sorted((min, value, max))[1]

    def available_pins(self):
        self._socket.send(b'LISTPORTS\n')
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            return json.loads(answer[3:])
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _set_pin_direction(self, pin, direction):
        direction = 'INPUT' if direction == ahio.Direction.Input else 'OUTPUT'
        command = ('SETDIRECTION %s %s\n' % (pin, direction)).encode('utf8')
        self._socket.send(command)
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            return None
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _pin_direction(self, pin):
        command = ('DIRECTION %s\n' % pin).encode('utf8')
        self._socket.send(command)
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            direction = answer[3:].strip()
            d = ahio.Direction
            return d.Input if direction == 'INPUT' else d.Output
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _set_pin_type(self, pin, ptype):
        ptype = 'DIGITAL' if ptype == ahio.PortType.Digital else 'ANALOG'
        command = ('SETTYPE %s %s\n' % (pin, ptype)).encode('utf8')
        self._socket.send(command)
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            return None
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _pin_type(self, pin):
        command = ('TYPE %s\n' % pin).encode('utf8')
        self._socket.send(command)
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            ptype = answer[3:].strip()
            pt = ahio.PortType
            return pt.Digital if ptype == 'DIGITAL' else pt.Analog
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _find_port_info(self, pin):
        ps = [p for p in self.available_pins() if p['id'] == pin]
        if ps:
            return ps[0]
        else:
            return None

    def _write(self, pin, value, pwm):
        if self._pin_direction(pin) == ahio.Direction.Input:
            return None
        pin_info = self._find_port_info(pin)
        if self._pin_type(pin) == ahio.PortType.Digital:
            if not pin_info['digital']['output']:
                raise RuntimeError('Pin does not support digital output')
            if pwm:
                if not pin_info['digital']['pwm']:
                    raise RuntimeError('Pin does not support PWM')
                value = self.__clamp(value, 0, 1)
                command = ('WRITEPWM %s %s\n' % (pin, value)).encode('utf8')
            else:
                value = 'HIGH' if value == ahio.LogicValue.High else 'LOW'
                command = ('WRITEDIGITAL %s %s\n' %
                           (pin, value)).encode('utf8')
        else:
            if not pin_info['analog']['output']:
                raise RuntimeError('Pin does not support analog output')
            l = pin_info['analog']['write_range']
            value = self.__clamp(value, l[0], l[1])
            command = ('WRITEANALOG %s %s\n' % (pin, value)).encode('utf8')
        self._socket.send(command)
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            return None
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _read(self, pin):
        pin_info = self._find_port_info(pin)
        pin_type = self._pin_type(pin)
        if pin_info['digital']['input'] and pin_type == ahio.PortType.Digital:
            da = ahio.PortType.Digital
            command = ('READDIGITAL %s\n' % pin).encode('utf8')
        elif pin_info['analog']['input'] and pin_type == ahio.PortType.Analog:
            da = ahio.PortType.Analog
            command = ('READANALOG %s\n' % pin).encode('utf8')
        else:
            raise RuntimeError('Pin does not support input or is not set up')
        self._socket.send(command)
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            value = answer[3:].strip()
            if da == ahio.PortType.Digital:
                lv = ahio.LogicValue
                return lv.High if value == 'HIGH' else lv.Low
            else:
                return int(value)
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def analog_references(self):
        self._socket.send(b'ANALOGREFERENCES\n')
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            __, *opts = answer.strip().split(' ')
            return opts
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _set_analog_reference(self, reference, pin):
        if pin:
            command = ('SETANALOGREFERENCE %s %s\n' % (reference, pin))
        else:
            command = ('SETANALOGREFERENCE %s\n' % reference)
        self._socket.send(command.encode('utf8'))
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            return
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _analog_reference(self, pin):
        if pin:
            command = 'ANALOGREFERENCE %s\n' % pin
        else:
            command = 'ANALOGREFERENCE\n'
        self._socket.send(command.encode('utf8'))
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            return answer.strip().split(' ')[1]
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')

    def _set_pwm_frequency(self, frequency, pin):
        if pin:
            command = 'SETPWMFREQUENCY %s %s\n' % (frequency, pin)
        else:
            command = 'SETPWMFREQUENCY %s\n' % frequency
        self._socket.send(command.encode('utf8'))
        with self._socket.makefile() as f:
            answer = f.readline()
        if answer.startswith('OK'):
            return
        elif answer.startswith('ERROR'):
            raise RuntimeError(answer[6:])
        else:
            raise RuntimeError('Unknown response')



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio.abstract_driver

import os.path
import platform
import re
from enum import Enum


def pi_version():
    """Detect the version of the Raspberry Pi.  Returns either 1, 2 or
    None depending on if it's a Raspberry Pi 1 (model A, B, A+, B+),
    Raspberry Pi 2 (model B+), or not a Raspberry Pi.
    https://github.com/adafruit/Adafruit_Python_GPIO/blob/master/Adafruit_GPIO/Platform.py
    """
    if not os.path.isfile('/proc/cpuinfo'):
        return None
    # Check /proc/cpuinfo for the Hardware field value.
    # 2708 is pi 1
    # 2709 is pi 2
    # Anything else is not a pi.
    with open('/proc/cpuinfo', 'r') as infile:
        cpuinfo = infile.read()
    # Match a line like 'Hardware   : BCM2709'
    match = re.search('^Hardware\s+:\s+(\w+)$', cpuinfo,
                      flags=re.MULTILINE | re.IGNORECASE)
    if not match:
        # Couldn't find the hardware, assume it isn't a pi.
        return None
    if match.group(1) == 'BCM2708':
        # Pi 1
        return 1
    elif match.group(1) == 'BCM2709':
        # Pi 2
        return 2
    elif match.group(1) == 'BCM2835':
        # Pi 3
        return 3
    else:
        # Something else, not a pi.
        return None


class ahioDriverInfo(ahio.abstract_driver.AbstractahioDriverInfo):
    NAME = 'Raspberry'
    AVAILABLE = pi_version() is not None


# Just try to import it. Code that depends on it should not be executed outside
# of Raspberry Pi anyway
try:
    import RPi.GPIO as GPIO
except ImportError:
    if ahioDriverInfo.AVAILABLE:
        print("RPi.GPIO not installed: sudo apt-get install python3-rpi.gpio")
        ahioDriverInfo.AVAILABLE = false
except RuntimeError:
    ahioDriverInfo.AVAILABLE = false
    print("You probably need superuser privileges. Try running with sudo")


class Driver(ahio.abstract_driver.AbstractDriver):

    Pins = Enum(
        'Pins',
        'D3 D5 D7 D8 D10 D12 D13 D15 D16 D18 D19 D21 D22 D23 D24 D26')

    __pwm = {}
    __pwm_frequency = {}

    def __init__(self):
        if ahioDriverInfo.AVAILABLE:
            GPIO.setmode(GPIO.BOARD)
            for pin in Driver.Pins:
                self._set_pin_direction(pin, ahio.Direction.Output)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        GPIO.cleanup()

    def __create_pin_info(self, pid):
        return {
            'id': pid,
            'name': 'Digital %s' % self.__pin_to_int(pid),
            'analog': {
                'input': False,
                'output': False,
                'read_range': None,
                'write_range': None
            },
            'digital': {
                'input': True,
                'output': True,
                'pwm': True
            }
        }

    def __pin_to_int(self, pin):
        if type(pin) is int:
            return pin
        else:
            return int(pin.name.replace('D', ''))

    def __clamp(self, value, min, max):
        return sorted((min, value, max))[1]

    # pinout used is the second on this link (physical/BOARD pinout):
    # https://www.raspberrypi.org/documentation/usage/gpio
    def available_pins(self):
        return [self.__create_pin_info(pin) for pin in Driver.Pins]

    def _set_pin_direction(self, pin, direction):
        pin = self.__pin_to_int(pin)
        if direction == ahio.Direction.Input:
            GPIO.setup(pin, GPIO.IN)
        else:
            GPIO.setup(pin, GPIO.OUT, initial=GPIO.LOW)
        self.__pwm.pop(pin, None)

    def _pin_direction(self, pin):
        pin = self.__pin_to_int(pin)
        function = GPIO.gpio_function(pin)
        if function == GPIO.IN:
            return ahio.Direction.Input
        elif function == GPIO.OUT or function == GPIO.PWM:
            return ahio.Direction.Output
        else:
            return None

    def _set_pin_type(self, pin, ptype):
        if ptype != ahio.PortType.Digital:
            raise RuntimeError('Raspberry Pi pins can only be used as Digital')

    def _pin_type(self, pin):
        return ahio.PortType.Digital

    def _write(self, pin, value, pwm):
        pin = self.__pin_to_int(pin)
        if self._pin_direction(pin) == ahio.Direction.Input:
            return
        if pwm:
            if type(value) is int or type(value) is float:
                value = int(100 * self.__clamp(float(value), 0.0, 1.0))
                p = self.__pwm.get(pin, None)
                if not p:
                    freq = self.__pwm_frequency.get(pin, None)
                    if not freq:
                        freq = 1
                    p = GPIO.PWM(pin, freq)
                p.start(value)
            else:
                raise TypeError(
                    'Value should be a float or int between 0 and 1')
        else:
            if type(value) is ahio.LogicValue:
                lv = ahio.LogicValue
                value = GPIO.HIGH if value == lv.High else GPIO.LOW
                GPIO.output(pin, value)
            else:
                raise TypeError('Value should be of type ahio.LogicValue')

    def _read(self, pin):
        pin = self.__pin_to_int(pin)
        return GPIO.input(pin)

    def analog_references(self):
        return []

    def _set_analog_reference(self, reference, pin):
        raise RuntimeError('Raspberry Pi does not have analog pins')

    def _analog_reference(self, pin):
        raise RuntimeError('Raspberry Pi does not have analog pins')

    def _set_pwm_frequency(self, frequency, pin):
        if pin:
            pin = self.__pin_to_int(pin)
            self.__pwm_frequency[pin] = frequency
            pwm = self.__pwm.get(pin, None)
            if pwm:
                pwm.ChangeFrequency(frequency)
        else:
            for pin in Driver.Pins:
                self._set_pwm_frequency(frequency, pin)



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from enum import Enum

import ahio.abstract_driver
import numpy as np
import scipy.signal


class ahioDriverInfo(ahio.abstract_driver.AbstractahioDriverInfo):
    NAME = 'SISO Model'
    AVAILABLE = True


class Driver(ahio.abstract_driver.AbstractDriver):
    Pins = Enum('Pins', 'U, Y')
    x = None
    u = None
    model = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def setup(self, model='([[0.5]], [[1]], [[1]], [[0]])'):
        self.model = [np.array(x) for x in eval(model)]

        if len(self.model) % 2 == 1:
            *self.model, dt = self.model
            dt = np.asscalar(dt)
            G = scipy.signal.dlti(*self.model, dt=dt)
            G = scipy.signal.StateSpace(G)
            self.model = G.A, G.B, G.C, G.D
        else:
            G = scipy.signal.lti(*self.model)
            G = scipy.signal.StateSpace(G)
            self.model = G.A, G.B, G.C, G.D
            vals = scipy.linalg.eigvals(G.A)
            dt = max(0.1, float('%.1f' % (max(abs(np.real(vals))) / 5)))
            *self.model, _ = scipy.signal.cont2discrete(self.model, dt)

        A, B, C, D = self.model
        self.u = 0
        self.x = np.zeros((A.shape[0], 1))

    def __create_pin_info(self, pid, pwm=False):
        obj = {
            'id': pid,
            'name': 'U' if pid.value == 1 else 'Y',
            'analog': {
                'input': pid.value == 2,
                'output': pid.value == 1,
                'read_range': (0, 100),
                'write_range': (0, 100)
            },
            'digital': {
                'input': False,
                'output': False,
                'pwm': False
            }
        }
        return obj

    def available_pins(self):
        pins = [p for p in Driver.Pins]
        return [self.__create_pin_info(pin, True) for pin in pins]

    def _set_pin_direction(self, pin, direction):
        pass

    def _pin_direction(self, pin):
        return ahio.Direction.Input if pin == 2 else ahio.Direction.Output

    def _set_pin_type(self, pin, ptype):
        pass

    def _pin_type(self, pin):
        return ahio.PortType.Digital

    def _write(self, pin, value, pwm):
        self.u = value

    def _read(self, pin):
        A, B, C, D = self.model
        if A.shape[0] == 1:
            self.x = A * self.x + B * self.u
            y = C * self.x
        else:
            self.x = A @ self.x + B * self.u
            y = C @ self.x
        return np.asscalar(y)

    def analog_references(self):
        return []

    def _set_analog_reference(self, reference, pin):
        pass

    def _analog_reference(self, pin):
        return []

    def _set_pwm_frequency(self, frequency, pin):
        pass



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""@package ahio
ahio provides access to various I/O devices.

This package provides abstracted access to various digital and analogic I/O
devices, such as Arduino, Raspberry PI, MyRIO and PLC. It allows the developer
to write and read data from such devices without needing to worry about the
specifics of each device, and to easily switch devices without code
refactoring.

The functions in this package list and instantiate the drivers. For the driver
api, see `ahio.abstract_driver.AbstractDriver`. Driver metadata format can be
found in `ahio.abstract_driver.AbstractahioDriverInfo`

@author Álan Crístoffer <acristoffers@gmail.com>
"""

from enum import Enum

import ahio.drivers

__author__ = 'Álan Crístoffer'
__copyright__ = 'Copyright 2016, Álan Crístoffer'
__credits__ = ['Álan Crístoffer']
__license__ = 'MIT'
__version__ = '1.0.14'
__maintainer__ = 'Álan Crístoffer'
__email__ = 'acristoffers@gmail.com'
__status__ = 'Release'

PortType = Enum('PortType', 'Analog Digital')
Direction = Enum('Direction', 'Output Input')
LogicValue = Enum('LogicValue', 'Low High')


def add_path(path):
    """Adds a directory to the list of folders where to load drivers from"""
    drivers.add_path(path)


def remove_path(path):
    """Removes a directory from the list of folders where to load drivers from"""
    drivers.remove_path(path)


def clear_path():
    """Clears the list of folders where to load drivers from"""
    drivers.clear_path()


def list_available_drivers():
    """Returns a list of string with the names of available drivers.

    Available means that the driver is installed and can be used. For example,
    it will not contain "Raspberry" if you're not running on a Raspberry Pi,
    even if the raspberry.py script is present in the drivers directory.

    @returns a list of strings that can be fed to `ahio.new_driver` to get an
    instance of the desired driver.
    """
    return drivers.available_drivers()


def driver_info(name):
    """Returns driver metadata.

    Returns a class which static properties contains metadata from the
    driver, such as name and availability.

    @returns a subclass from `ahio.abstract_driver.AbstractahioDriverInfo` with
    metadata from the driver.
    """
    return drivers.driver_info(name)


def new_driver(name):
    """Instantiates a new object of the named driver.

    The API used by the returned object can be seen in
    `ahio.abstract_driver.AbstractDriver`

    @returns a Driver object from the required type or None if it's not
    available
    """
    return drivers.new_driver_object(name)



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import importlib.machinery
import os
import sys

# path to the drivers folder
__modules_path = os.path.dirname(os.path.realpath(__file__))
# additional added paths
__modules_path_user = []
# tries to load every file in the drivers folder
__modules = (__load_driver(__modules_path + '/' + driver)
             for driver in os.listdir(__modules_path))
# filters out the False elements, leaving only valid drivers
__modules = (d for d in __modules if d)
# only installed drivers that are available in this platform
__available = None
# counts the number of drivers loaded. The number is appended to driver name
# to avoid collision
__count = 0


def add_path(path):
    global __modules_path_user
    global __modules
    global __available
    __modules_path_user.append(path)
    path = [__modules_path, *__modules_path_user]
    fs = [d + '/' + f for d in path for f in os.listdir(d)]
    __modules = (__load_driver(driver) for driver in fs)
    __modules = (d for d in __modules if d)
    __available = None


def remove_path(path):
    global __modules_path_user
    global __modules
    global __available
    __modules_path_user = [x for x in __modules_path_user if x != path]
    path = [__modules_path, *__modules_path_user]
    fs = [d + '/' + f for d in path for f in os.listdir(d)]
    __modules = (__load_driver(driver) for driver in fs)
    __modules = (d for d in __modules if d)
    __available = None


def clear_path():
    global __modules_path_user
    global __modules
    global __available
    __modules_path_user = []
    fs = [d + '/' + f for d in [__modules_path] for f in os.listdir(d)]
    __modules = (__load_driver(driver) for driver in fs)
    __modules = (d for d in __modules if d)
    __available = None


def available_drivers():
    """Returns a list of available drivers names.
    """
    global __modules
    global __available

    if type(__modules) is not list:
        __modules = list(__modules)

    if not __available:
        __available = [d.ahioDriverInfo.NAME
                       for d in __modules
                       if d.ahioDriverInfo.AVAILABLE]

    return __available


def driver_info(name):
    """Returns driver metadata.

    Returns a class which static properties contains metadata from the
    driver, such as name and availability.

    @returns a subclass from `ahio.abstract_driver.AbstractahioDriverInfo` with
    metadata from the driver.
    """
    driver = __locate_driver_named(name)
    return driver.ahioDriverInfo if driver else None


def new_driver_object(name):
    """Instantiates a new object of the named driver.

    The API used by the returned object can be seen in
    `ahio.abstract_driver.AbstractDriver`

    @returns a Driver object from the required type of None if it's not
    available
    """
    driver = __locate_driver_named(name)
    return driver.Driver() if driver else None


def __load_driver(name):
    """Tries to load the driver named @arg name.

    A driver is considered valid if it has a ahioDriverInfo object. It should
    however implement all APIs described in `ahio.abstract_driver`, as they'll
    be needed to use the driver.

    @returns the driver package, or False if it failed.
    """
    global __count
    try:
        dname = os.path.basename(name).replace('.py', '')
        mod_name = 'ahio.drivers.%s%d' % (dname, __count)
        loader = importlib.machinery.SourceFileLoader(mod_name, name)
        driver = loader.load_module()
        __count += 1
        return driver if hasattr(driver, 'ahioDriverInfo') else False
    except Exception:
        return False


def __locate_driver_named(name):
    """Searchs __modules for a driver named @arg name.

    @returns the package for driver @arg name or None if one can't be found.
    """
    global __modules

    if type(__modules) is not list:
        __modules = list(__modules)

    ms = [d for d in __modules if d.ahioDriverInfo.NAME == name]
    if not ms:
        return None
    return ms[0]



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio.abstract_driver

import serial
import time
from enum import Enum


class ahioDriverInfo(ahio.abstract_driver.AbstractahioDriverInfo):
    NAME = 'Arduino'
    AVAILABLE = True


class Driver(ahio.abstract_driver.AbstractDriver):
    _serial = None

    Pins = Enum(
        'Pins',
        'D0 D1 D2 D3 D4 D5 D6 D7 D8 D9 D10 D11 D12 D13 A0 A1 A2 A3 A4 A5')

    AnalogReferences = Enum('AnalogReferences', 'Default Internal External')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def setup(self, port):
        """Connects to an Arduino UNO on serial port `port`.

        @throw RuntimeError can't connect to Arduino
        """
        port = str(port)
        # timeout is used by all I/O operations
        self._serial = serial.Serial(port, 115200, timeout=2)
        time.sleep(2)  # time to Arduino reset

        if not self._serial.is_open:
            raise RuntimeError('Could not connect to Arduino')

        self._serial.write(b'\x01')

        if self._serial.read() != b'\x06':
            raise RuntimeError('Could not connect to Arduino')

        ps = [p for p in self.available_pins() if p['digital']['output']]
        for pin in ps:
            self._set_pin_direction(pin['id'], ahio.Direction.Output)

    def __clamp(self, value, min, max):
        return sorted((min, value, max))[1]

    def __create_pin_info(self, pid, pwm=False):
        is_analog = pid.name.startswith('A')
        obj = {
            'id': pid,
            'name': None,
            'analog': {
                'input': is_analog,
                'output': False,
                'read_range': (0, 1023) if is_analog else None,
                'write_range': None
            },
            'digital': {
                'input': not is_analog,
                'output': not is_analog,
                'pwm': not is_analog and pwm
            }
        }
        if is_analog:
            obj['name'] = 'Analog %s' % (pid.value - 14)
        else:
            obj['name'] = 'Digital %s' % (pid.value - 1)
        return obj

    def available_pins(self):
        pins = [p for p in Driver.Pins]
        pwms = [self.__create_pin_info(pin, True)
                for pin in pins
                if (pin.value - 1) in [3, 5, 6, 9, 10, 11]]
        pins = [self.__create_pin_info(pin)
                for pin in pins
                if (pin.value - 1) not in [3, 5, 6, 9, 10, 11]]
        return sorted(pwms + pins, key=lambda pin: pin['id'].value)

    def _set_pin_direction(self, pin, direction):
        if pin.name.startswith('A') and direction == ahio.Direction.Output:
            raise RuntimeError('Analog pins can only be used as Input')
        if direction == ahio.Direction.Input:
            self._serial.write(
                b'\x02\xC3' + bytes({pin.value - 1}) + bytes({1}))
        else:
            self._serial.write(
                b'\x02\xC3' + bytes({pin.value - 1}) + bytes({0}))
            self._serial.write(
                b'\x02\xC7' + bytes({pin.value - 1}) + bytes({0}))

    def _pin_direction(self, pin):
        self._serial.write(b'\x02\xC4' + bytes({pin.value - 1}))
        direction = self._serial.read()
        if direction == b'\x01':
            return ahio.Direction.Input
        elif direction == b'\x00':
            return ahio.Direction.Output
        else:
            return None

    def _set_pin_type(self, pin, ptype):
        is_analog = pin.name.startswith('A')
        if is_analog and ptype == ahio.PortType.Digital:
            raise RuntimeError('Analog pin can not be set as digital')
        if not is_analog and ptype == ahio.PortType.Analog:
            raise RuntimeError('Digtal pin can not be set as analog')

    def _pin_type(self, pin):
        pt = ahio.PortType
        return pt.Analog if pin.name.startswith('A') else pt.Digital

    def _write(self, pin, value, pwm):
        if self._pin_direction(pin) == ahio.Direction.Input:
            return
        if pin.name.startswith('D'):
            if pwm:
                if (pin.value - 1) not in [3, 5, 6, 9, 10, 11]:
                    raise RuntimeError('Pin does not support PWM output')
                if type(value) is int or type(value) is float:
                    value = int(255 * self.__clamp(float(value), 0.0, 1.0))
                    command = b'\x02\xC8' + bytes({pin.value - 1})
                    arg = bytes({value})
                    self._serial.write(command + arg)
                else:
                    raise TypeError('value not a float or int between 0 and 1')
            else:
                if type(value) is ahio.LogicValue:
                    value = 1 if value == ahio.LogicValue.High else 0
                    command = b'\x02\xC7' + bytes({pin.value - 1})
                    arg = bytes({value})
                    self._serial.write(command + arg)
                else:
                    raise TypeError('Value should be of type ahio.LogicValue')
        else:
            raise RuntimeError('Can not write to analog pin')

    def _read(self, pin):
        if pin.name.startswith('D'):
            self._serial.write(b'\x02\xC5' + bytes({pin.value - 1}))
            value = self._serial.read()
            lv = ahio.LogicValue
            return lv.High if value == b'\x01' else lv.Low
        else:
            self._serial.write(b'\x02\xC6' + bytes({pin.value - 14}))
            value_high = self._serial.read()
            value_low = self._serial.read()
            return (value_high[0] << 8) | value_low[0]

    def analog_references(self):
        return [r for r in Driver.AnalogReferences]

    def _set_analog_reference(self, reference, pin):
        if pin is not None:
            raise RuntimeError('Per pin analog reference is not supported')
        self._serial.write(b'\x02\xC2' + bytes({reference.value - 1}))

    def _analog_reference(self, pin):
        self._serial.write(b'\x02\xC1')
        reference = self._serial.read()[0]
        return [Driver.AnalogReferences.Default,
                Driver.AnalogReferences.Internal,
                Driver.AnalogReferences.External][reference]

    def _set_pwm_frequency(self, frequency, pin):
        raise RuntimeError(
            'Setting PWM frequency is not supported by hardware')
